/*
Turing Machine Simulator
Copyright (C) 2005-2011 Samuel BWPOW Kupka, http://bwpow.eu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <string>
#include <vector>
#include <map>
#include "ts.h"

using namespace std;

bool operator<(const INSTR &a,const INSTR &b)
{
  if(a.s==b.s) return a.p<b.p;
  return a.s<b.s;
}

map<INSTR,INSTR> instr;
vector<string> state;
vector<char> pr,pl;
INFO info;
char chars[256];

const char *license[]={
"This program is free software: you can redistribute it and/or modify",
"it under the terms of the GNU General Public License as published by",
"the Free Software Foundation, either version 3 of the License, or",
"(at your option) any later version.",
"",
"This program is distributed in the hope that it will be useful,",
"but WITHOUT ANY WARRANTY; without even the implied warranty of",
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the",
"GNU General Public License for more details.",
"",
"You should have received a copy of the GNU General Public License",
"along with this program.  If not, see http://www.gnu.org/licenses/.",
NULL
};

void main_help(void)
{
  printf("Turing Machine Simulator %s\n",VERSION);
  printf("Copyright (C) 2005-2011 Samuel BWPOW Kupka, http://bwpow.eu\n");
  printf("Created for PALMA.STROM.SK project (http://palma.strom.sk)\n");
  printf("============================================================================\n\n");
  printf("Usage: palma_ts.exe tsfile inputfile outputfile maxlength maxsteps [options]\n\n");
  printf("Options:\n");
  printf("-t0 : (default) Machine without constraints\n");
  printf("-t1 : Left constraint\n");
  printf("-t2 : Left and right constraint (maximal length is used as length)\n");
  printf("-nt : Enable constraint characters. Otherwise machine will just stop after leaving existing tape\n");
  printf("-ic : Create unary encoded numbers consisting of ones from numbers on input\n");
  printf("-ics: Create unary encoded numbers consisting of ones from numbers on input, but leave one zero at the beginning of the tape\n");
  printf("-ot : Output tape\n");
  printf("-os : Output state\n");
  printf("-oc : Output count of ones (second character in input file)\n");
  printf("-op : Output relative position of the head\n");
  printf("-tle: Consider Time and Space limit as regular end of simulation\n");
  printf("-sp : Set starting position to the first one (second character in input file)\n");
  printf("These options can be placed at the beginning of the input file.\n");
  printf("\n");
  for(int i=0;license[i];i++) printf("%s\n",license[i]);
  exit(1);
}

void main_exit(const char *text)
{
  if(strlen(text)>0) printf("%s\n",text);
  exit(1);
}

int main(int argc,char *argv[])
{
  int r;
  if(argc<6) main_help();

  memset(&info,0,sizeof(INFO));

  for(int i=6;i<argc;i++){
    if(!stricmp(argv[i],"-t0")){ info.cl=0;info.cr=0; }
    if(!stricmp(argv[i],"-t1")){ info.cl=1;info.cr=0; }
    if(!stricmp(argv[i],"-t2")){ info.cl=1;info.cr=1; }
    if(!stricmp(argv[i],"-nt")){ info.nt=1; }
    if(!stricmp(argv[i],"-ot")){ info.output|=OUTPUT_CONF; }
    if(!stricmp(argv[i],"-os")){ info.output|=OUTPUT_STATE; }
    if(!stricmp(argv[i],"-oc")){ info.output|=OUTPUT_COUNT; }
    if(!stricmp(argv[i],"-op")){ info.output|=OUTPUT_POS; }
    if(!stricmp(argv[i],"-ic")){ info.input=1; }
    if(!stricmp(argv[i],"-ics")){ info.input=2; }
    if(!stricmp(argv[i],"-tle")){ info.tle=1; }
    if(!stricmp(argv[i],"-sp")){ info.spos=1; }
  }
  strcpy(info.tsfile,argv[1]);
  strcpy(info.infile,argv[2]);
  strcpy(info.outfile,argv[3]);
  if(sscanf(argv[4],"%ld",&info.maxmemory)!=1) main_help();
  if(sscanf(argv[5],"%ld",&info.maxsteps)!=1) main_help();

  if(load_input()!=0) main_exit("Error loading input file!");
  r=load_ts();
  if(r==-1) main_exit("Error opening TS file!");
  if(r==-2) main_exit("Too many states.");
  if(r==-3) main_exit("Too many instructions.");
  if(r==-4) main_exit("Invalid character.");
  if(r==-5) main_exit("Non-deterministic machine.");
  if(r==-6) main_exit("Unknown movement.");

  r=run();
  if(info.tle==0&&r==-1) main_exit("Time limit exceeded.");
  if(info.tle==0&&r==-2) main_exit("Memory limit exceeded.");

  if(info.tle&&(r==-1||r==-2)){
    if(save_tle_output()!=0) main_exit("Can't write to output file!");
  }
  else{
    if(save_output()!=0) main_exit("Can't write to output file!");
  }

  print_output();
  return 0;
}
